namespace SoMRandomizer.processing.openworld.randomization
{
    /// <summary>
    /// Unique IDs of PrizeLocations for use with multiworld. They are roughly in order of definition.
    /// Do not change or reorder. Only append or fill holes.
    /// </summary>
    public enum LocationId : byte
    {
        /// <summary>Pseudo Location</summary>
        StarterWeaponMain = 1,
        /// <summary>Pseudo Location</summary>
        StarterWeaponAlt = 2,
        /// <summary>Reserved, unused</summary>
        StarterWeaponAlt2 = 3,
        MechRider3 = 4,
        Buffy = 5,
        DreadSlime = 6,
        GnomeItem1 = 7,
        GnomeItem2 = 8,
        FireSeed = 9,
        LunaItem1 = 10,
        LunaItem2 = 11,
        Kakkara = 12,
        LuminaSpells = 13,
        LuminaSeed = 14,
        ChestNextToWhipChest = 15,
        ShadePalaceGloveOrbChest = 16,
        SunkenContinentSwordOrbChest = 17,
        LuminaTowerAxeOrbChest = 18,
        FirePalaceAxeOrbChest = 19,
        LuminaTowerSpearOrbChest = 20,
        SunkenContinentBoomerangOrbChest = 21,
        FirePalaceChest1 = 22,
        FirePalaceChest2 = 23,
        Santa = 24,
        ShadeSpells = 25,
        ShadeSeed = 26,
        ThunderGigas = 27,
        RedDragon = 28,
        BlueDragon = 29,
        ManaTree = 30,
        MatangoFlammie = 31,
        Jehk = 32,
        Hydra = 33,
        Kettlekin = 34,
        ShadePalaceChest = 35,
        LukaItem1 = 36,
        LukaItem2 = 37,
        SylphidItem1 = 38,
        SylphidItem2 = 39,
        WhipChest = 40,
        MoogleVillageGloveOrbChest = 41,
        IceCastleGloveOrbChest = 42,
        PandoraSwordOrbChest = 43,
        NtRuinsSwordOrbChest = 44,
        MoogleVillageAxeOrbChest = 45,
        NtCastleAxeOrbChest = 46,
        NtRuinsSpearOrbChest = 47,
        PandoraSpearOrbChest = 48,
        SantaSpearOrbChest = 49,
        KilroyWhipOrbChest = 50,
        NtCastleWhipOrbChest = 51,
        NtRuinsBowOrbChest = 52,
        PotosChest = 53,
        PandoraChest1 = 54,
        PandoraChest2 = 55,
        PandoraChest3 = 56,
        PandoraChest4 = 57,
        MagicRopeChest = 58,
        NtCastleChest = 59,
        MatangoInnJavelinOrbChest = 60,
        Watts = 61,
        UndineItem1 = 62,
        UndineItem2 = 63,
        Salamando = 64,
        DryadSpells = 65,
        DryadSeed = 66,
        Mara = 67,
        TurtleIsland = 68,
        DwarfElder = 69,
        TropicalloSprite = 70,
        TropicalloBow = 71,
        Girl = 72,
        Lighthouse = 73,
        Kilroy = 74,
        MantisAnt = 75,
        AxeBeak = 76,
        SnowDragon = 77,
        DragonWorm = 78,
        DoomWall = 79,
        Vampire = 80,
        MechRider2 = 81,
        Watermelon = 82,
        Hexas = 83,
        WallFace = 84,
        MetalMantis = 85,
        JemaAtTasnica = 86,
        TripleTonpole = 87,
        SwordPedestal = 88,
        SantaGift0 = 89,
        SantaGift1 = 90,
        SantaGift2 = 91,
        SantaGift3 = 92,
        SantaGift4 = 93,
        SantaGift5 = 94,
        SantaGift6 = 95,
        SantaGift7 = 96,
    }
}
