
namespace SoMRandomizer.processing.ancientcave
{
    /// <summary>
    /// A bunch of Mitch Hedberg quotes organized in a way that SoM can display them.
    /// </summary>
    /// 
    /// <remarks>Author: Moppleton</remarks>
    public class MitchHedbergQuotes
    {
        // h4rdycore : " You know ... I find that a Rabite's opinion of me is influenced by whether or not I have Candy." - Mitch Hedburg
        // 28 characters max per line
        public static string[] mitchQuotes =
         {
            //                          /
            "This shirt is [dry-clean\n" +
            "only]...which means it's\n" +
            "dirty.",

            //                          /
            "If carrots got you drunk,\n" +
            "rabbits would be fucked up.",

            //                          /
            "I wish I could play little\n" +
            "league now, I'd kick some\n" +
            "fuckin' ass.",

            //                          /
            "My fake plants died because\n" +
            "I did not pretend to water\n" +
            "them.",

            //                          /
            "I don't have a microwave\n" +
            "oven, but I do have a clock\n" +
            "that occasionally\n" +
            "cooks shit.",

            //                          /
            "I had a parrot. The parrot\n" +
            "talked, but it did not say\n" +
            "[I'm hungry,] so it died.",

            //                          /
            "I order the club sandwich\n" +
            "all the time, but I'm not\n" +
            "even a member, man.\n" +
            "I don't know how I get away\n" +
            "with it.",

            //                          /
            "I saw this wino, he was\n" +
            "eating grapes. I was like,\n" +
            "[Dude, you have to wait.]",

            //                          /
            "I'd like to make a vending\n" +
            "machine that sells vending\n" +
            "machines.\n" +
            "It'd have to be real\n" +
            "fuckin' big!",

            //                          /
            "I'm against picketing, but\n" +
            "I don't know how to show it.",

            //                          /
            "I haven't slept for ten\n" +
            "days, because that would\n" +
            "be too long.",

            //                          /
            "My friend asked me if I\n" +
            "wanted a frozen banana,\n" +
            "I said\n" +
            "[No, but I want a regular\n" +
            "banana later, so ... yeah].",

            //                          /
            "I used to do drugs.\n" +
            "I still do, but I used to,\n" +
            "too.",

            //                          /
            "I bought a seven-dollar\n" +
            "pen because I always lose\n" +
            "pens and I got sick of\n" +
            "not caring.",

            //                          /
            "I like rice. Rice is great\n" +
            "when you're hungry and you\n" +
            "want 2,000 of something.",

            //                          /
            "I love my fed-ex guy cause\n" +
            "he's a drug dealer and he\n" +
            "doesn't even know it.",

            //                          /
            "I got a king sized bed. I\n" +
            "don't know any kings, but\n" +
            "if one came over,\n" +
            "I guess he'd be\n" +
            "comfortable.",

            //                          /
            "You know that word [lull]?\n" +
            "That's four letters, three\n" +
            "of them are L's, fuck!\n" +
            "That's two many L's in one\n" +
            "word!",

            //                          /
            "They say Flintstone's\n" +
            "vitamins are chewable. All\n" +
            "vitamins are chewable,\n" +
            "it's just that they taste\n" +
            "shitty.",

            //                          /
            "I went to a record store,\n" +
            "they said they specialized\n" +
            "in hard-to-find records.\n" +
            "Nothing was alphabetized.",

            //                          /
            "Imagine if the headless\n" +
            "horseman had a headless\n" +
            "horse.\n" +
            "That would be fucking chaos.",

            //                          /
            "A fly was very close to\n" +
            "being called a [land,]\n" +
            "cause that's what they do\n" +
            "half the time.",

            //                          /
            "My belt holds up my pants\n" +
            "and my pants have belt\n" +
            "loops that hold up\n" +
            "the belt.  What the fuck's\n" +
            "really goin' on down there?\n" +
            "Who is the real hero?",

            //                          /
            "I can't tell you what hotel\n" +
            "I'm stayin' in, but there\n" +
            "are two trees involved.",

            //                          /
            "Fish are always eating\n" +
            "other fish. If fish could\n" +
            "scream, the ocean would be\n" +
            "loud as shit.",

            //                          /
            "I like vending machines\n" +
            "'cause snacks are better\n" +
            "when they fall.",

            //                          /
            "This guy handed me a\n" +
            "picture of him. He said,\n" +
            "[Here's a picture of me\n" +
            "when I was younger.] Every\n" +
            "picture is of you when\n" +
            "you were younger.",

            //                          /
            "Snake eyes!  That's a\n" +
            "gambling term.  Or it's an\n" +
            "animal term too.",

            //                          /
            "I think Bigfoot is blurry,\n" +
            "that's the problem. It's\n" +
            "not the photographer's\n" +
            "fault.",

            //                          /
            "I was walking down the\n" +
            "street with my friend and\n" +
            "he said,\n" +
            "[I hear music], as if\n" +
            "there is any other way you\n" +
            "can take it in.",

            //                          /
            "I'm sick of following my\n" +
            "dreams, man. I'm just gonna\n" +
            "ask where they're going and\n" +
            "hook up with 'em later.",

            //                          /
            "I bought a donut and they\n" +
            "gave me a receipt for the\n" +
            "donut.\n" +
            "I don't need a receipt for\n" +
            "the donut.  I'll just give\n" +
            "you the money and you give\n" +
            "me the donut.  End of\n" +
            "transaction. We don't need\n" +
            "to bring ink and paper\n" +
            "into this.",

            //                          /
            "You know they call corn on\n" +
            "the cob [corn on the cob,]\n" +
            "right? But that's how it\n" +
            "comes out of the ground,\n" +
            "man. They should call that\n" +
            "[corn,] and they should\n" +
            "call every other version\n" +
            "[corn off the cob.]",

            //                          /
            "I hope next time I move I\n" +
            "get a real easy phone\n" +
            "number, something that's\n" +
            "real easy to remember.\n" +
            "Something like two two two\n" +
            "two two two two.  I would\n" +
            "say [Sweet.] And then\n" +
            "people would say [Mitch,\n" +
            "how do I get ahold of you?]\n" +
            "I'd say [Just press two for\n" +
            "a while. And when I answer,\n" +
            "you will know you have\n" +
            "pressed two enough.",
            
            //                          /
            "This is what my friend said\n" +
            "to me, he said [You know\n" +
            "what I like? Mashed\n" +
            "potatoes.] It's like, dude,\n" +
            "you gotta give me time to\n" +
            "guess. If you're gonna quiz\n" +
            "me, you must insert a pause\n" +
            "in there.",

            //                          /
            "Last week I helped my\n" +
            "friend stay put. It's a lot\n" +
            "easier than helpin' him\n" +
            "move. I just went over to\n" +
            "his house and made sure he\n" +
            "did not start to load shit\n" +
            "into a truck.",

            //                          /
            "Whenever I go to shave, I\n" +
            "assume there's someone else\n" +
            "on the planet shaving, so I\n" +
            "say [I'm gonna go shave,\n" +
            "too.]",

            //                          /
            "On a traffic light, green\n" +
            "means [go,] and yellow\n" +
            "means [yield,] but on a\n" +
            "banana it's just the\n" +
            "opposite. Green means\n" +
            "[hold on,] yellow means\n" +
            "[go ahead,] and red means\n" +
            "[where the fuck did you\n" +
            "get that banana at?]",

            //                          /
            "My friend said to me, [I\n" +
            "think the weather's\n" +
            "trippy.] I said, [No man,\n" +
            "it's not the weather that\n" +
            "is trippy, perhaps it is\n" +
            "the way we perceive it\n" +
            "that is indeed trippy.]\n" +
            "Then I thought, man, I\n" +
            "should have just said\n" +
            "[Yeah.]",

            //                          /
            "Fettucini alfredo is\n" +
            "macaroni and cheese for\n" +
            "adults.",

            //                          /
            "I wrote a script and gave\n" +
            "it to a guy that reads\n" +
            "scripts, and he read it\n" +
            "and said he really likes\n" +
            "it, but he thinks I need\n" +
            "to rewrite it.  I said,\n" +
            "[Fuck that, I'll just\n" +
            "make a copy.]",

            //                          /
            "I was walking by a dry\n" +
            "cleaner at 3 AM and it\n" +
            "said, [Sorry, we're\n" +
            "closed.] You don't have\n" +
            "to be sorry. It's 3 AM\n" +
            "and you are a dry cleaner.\n" +
            "It would be ridiculous for\n" +
            "me to expect you to be\n" +
            "open.",

            //                          /
            "An escalator can never\n" +
            "break, it can only become\n" +
            "stairs. You would never\n" +
            "see an [Escalator\n" +
            "temporarily out of order]\n" +
            "sign, just [Escalator\n" +
            "temporarily stairs.]\n" +
            "Sorry for the convenience.",

            //                          /
            "I went to the Home Depot\n" +
            "yesterday, which was\n" +
            "unnecessary. I need to go\n" +
            "to the Apartment Depot.\n" +
            "It's just a bunch of guys\n" +
            "standing around going,\n" +
            "[We ain't gotta fix shit.]",

            //                          /
            "I saw this commercial on\n" +
            "late night TV, it said\n" +
            "[Forget everything you\n" +
            "know about slipcovers.]\n" +
            "So I did. And it was a\n" +
            "load off my mind. Then\n" +
            "the commercial tried to\n" +
            "sell me slipcovers, but\n" +
            "I didn't know what the\n" +
            "hell they were.",
        };
    }
}
