
namespace SoMRandomizer.processing.ancientcave
{
    /// <summary>
    /// A collection of Demitri Martin quotes for ancient cave NPCs to say.
    /// </summary>
    /// 
    /// <remarks>Author: Moppleton</remarks>
    public class DemetriMartinQuotes
    {
        // http://funnycomedianquotes.com/funny-demetri-martin-jokes-and-quotes.html
        // https://www.goodreads.com/author/quotes/4394776.Demetri_Martin
        public static string[] quotes = new string[]
        {
            //                          /
            "I was on the street. This\n" +
            "guy waved to me, and he\n" +
            "came up to me and said,\n" +
            "[I'm sorry, I thought you\n" +
            "were someone else.]\n" +
            "And I said, [I am.]",

            //                          /
            "Employee of the month is\n" +
            "a good example of how\n" +
            "somebody can be both a\n" +
            "winner and a loser at\n" +
            "the same time.",

            //                          /
            "I wrapped my Christmas\n" +
            "presents early this year,\n" +
            "but I used the wrong paper.\n" +
            "See, the paper I used said\n" +
            "[Happy Birthday] on it. I\n" +
            "didn't want to waste it so\n" +
            "I just wrote [Jesus] on it.",

            //                          /
            "[Sort of] is such a\n" + 
            "harmless thing to say...\n" + 
            "sort of.\n" + 
            "It's just a filler. [Sort\n" + 
            "of.] It doesn't really\n" + 
            "mean anything.\n" + 
            "But after certain things,\n" + 
            "[sort of] means\n" + 
            "everything, like after\n" + 
            "[I love you,] or\n" + 
            "\n" +
            "\n" +
            "[You're going to live], or\n" +
            "\n" +
            "\n" +
            "[It's a boy!]",

            //                          /
            "I feel stupid when I write\n" + 
            "the word banana. Its like,\n" +
            "\n" +
            "How many na's are on this\n" + 
            "thing?\n" +
            "\n" +
            "Cause I'm like\n" + 
            "[Bana... keep going.\n" + 
            "  Bananana... damn.]",

            //                          /
            "I was making pancakes the\n" + 
            "other day and a fly flew\n" + 
            "into the kitchen.\n" + 
            "And that's when I realized\n" + 
            "that a spatula is a lot\n" + 
            "like a fly swatter.\n" + 
            "And a crushed fly is a lot\n" + 
            "like a blueberry.\n" +
            "\n" + 
            "And a roommate is a lot\n" + 
            "like a fly eater.",

            //                          /
            "My plumbing is all screwed\n" + 
            "up. Because it turns out,\n" +
            "\n" + 
            "I do not own a\n" + 
            "garbage disposal.",

            //                          /
            "A lot of people don't like\n" + 
            "bumper stickers. I don't\n" + 
            "mind bumper stickers.\n" + 
            "To me a bumper sticker is\n" + 
            "a shortcut.\n" +
            "\n" +
            "It's like a little sign\n" + 
            "that says [Hey, let's\n" + 
            "never hang out.]",

            //                          /
            "A drunk driver is very\n" + 
            "dangerous. So is a drunk\n" + 
            "backseat driver, if he's\n" + 
            "persuasive.\n" + 
            "[Dude, make a left.]\n" +
            "\n" +
            "[Those are trees.]\n" +
            "\n" +
            "\n" +
            "[Trust me.]",

            //                          /
            "If I have to move up in a\n" + 
            "building, I choose the\n" + 
            "elevator over the\n" + 
            "escalator. Because one time\n" + 
            "I was riding the escalator\n" + 
            "and I tripped.\n" + 
            "I fell down the stairs for\n" + 
            "an hour and a half.",

            //                          /
            "Saying [I'm sorry] is the\n" + 
            "same as saying\n" + 
            "[I apologize.]\n" + 
            "Except at a funeral.",

            //                          /
            "I used to play sports.\n" + 
            "Then I realized you can\n" + 
            "buy trophies.\n" + 
            "Now I'm good at everything.",

            //                          /
            "I went into a clothes\n" + 
            "store and a lady came up\n" + 
            "to me and said\n" + 
            "[If you need anything,\n" + 
            "I'm Jill].\n" +
            "\n" +
            "I've never met anyone\n" + 
            "with a conditional\n" + 
            "identity before.",

            //                          /
            "About a month ago I got\n" + 
            "a cactus. A week later,\n" + 
            "it died.\n" + 
            "I was really depressed\n" + 
            "because I was like\n" +
            "\n" +
            "[Damn! I am less\n" + 
            "nurturing than a desert.]",

            //                          /
            "I like to use [I Can't\n" + 
            "Believe it's Not Butter] on\n" + 
            "my toast in the morning,\n" + 
            "because sometimes when I\n" + 
            "eat breakfast, I like to\n" + 
            "be incredulous.\n" + 
            "How was breakfast?\n" + 
            "Unbelievable.",

            //                          /
            "I wanna make a jigsaw\n" + 
            "puzzle that's 40,000\n" + 
            "pieces.\n" + 
            "And when you finish it,\n" + 
            "it says [go outside.]",

            //                          /
            "If you want to make a\n" + 
            "mythical creature, just\n" + 
            "take a regular animal and\n" + 
            "add wings to it.\n" + 
            "A horse becomes a Pegasus,\n" + 
            "a lion becomes a griffin,\n" + 
            "and a hawk...\n" + 
            "becomes a double hawk.",

            //                          /
            "When you have a fat friend\n" + 
            "there are no see-saws,\n" + 
            "only catapults.",

            //                          /
            "I keep a lighter in my\n" + 
            "back pocket all the time.\n" + 
            "I'm not a smoker,\n" + 
            "I just really like certain\n" + 
            "songs.",

            //                          /
            "I was in a card store and\n" + 
            "there were these cards\n" + 
            "that said [Get well soon.]\n" + 
            "Fuck that! Get well NOW.",

            //                          /
            "It's weird the way [finger\n" + 
            "puppet] sounds okay as a\n" + 
            "noun.",

            //                          /
            "Whenever I'm on my\n" + 
            "computer, I don't type\n" + 
            "[lol.]\n" + 
            "I type [lqtm]:\n" + 
            "[laugh quietly to myself.]\n" + 
            "It's more honest.",

            //                          /
            "If I ever saw an amputee\n" + 
            "being hanged, I would just\n" + 
            "yell out letters.",

            //                          /
            "There's a store in my\n" + 
            "neighborhood called\n" + 
            "Futon World.\n" + 
            "I like that name,\n" + 
            "[Futon World.] Makes me\n" + 
            "think of a magical place\n" + 
            "that gets less and less\n" + 
            "comfortable over time.",

            //                          /
            "I think they named oranges\n" + 
            "before they named carrots.\n" +
            "\n" +
            "[What are those?]\n" + 
            "[Those are orange...\n" + 
            "oranges.]\n" + 
            "[And what are those?]\n" +
            "\n" +
            "\n" +
            "[Ah, shit. Tall pointies?\n" + 
            "Are we going by shapes now?]",

            //                          /
            "I saw a transvestite\n" + 
            "wearing a T-shirt that said\n" + 
            "[Guess.]",

            //                          /
            "The easiest time to add\n" + 
            "insult to injury is when\n" + 
            "you're signing somebody's\n" + 
            "cast. [You're a dick! You\n" + 
            "deserved this! Also I'm\n" + 
            "sorry I broke your leg.]",

            //                          /
            "I wonder what the most\n" + 
            "intelligent thing ever\n" + 
            "said was that started with\n" + 
            "the word [dude.]\n" + 
            "[Dude, these are isotopes.]\n" +
            "\n" +
            "[Dude, we removed your\n" + 
            "kidney. You're going to\n" + 
            "be fine.]\n" + 
            "[Dude, I am so stoked to\n" + 
            "win this Nobel Prize. I\n" + 
            "just want to thank Kevin,\n" + 
            "Turtle, and all my homies.]",

            //                          /
            "I saw a guy at a party\n" + 
            "wearing a leather jacket\n" + 
            "and I thought,\n" + 
            "[That is cool.] But then I\n" + 
            "saw another guy wearing a\n" + 
            "leather vest and I thought,\n" + 
            "[That is not cool]. Then I\n" + 
            "figured it out:\n" +
            "\n" +
            "Cool is all about leather\n" + 
            "sleeves.",

            //                          /
            "The worst time to have a\n" + 
            "heart attack is during a\n" + 
            "game of charades.",

            //                          /
            "A couple weeks ago I was on\n" + 
            "the street and I saw an\n" + 
            "ugly pregnant lady,\n" + 
            "and I just thought,\n" + 
            "[Good for you.]",

            //                          /
            "The thing about glitter is\n" + 
            "if you get it on you, be\n" + 
            "prepared to have it on you\n" + 
            "forever. Glitter is the\n" + 
            "herpes of craft supplies.",

            //                          /
            "Whenever I see an\n" + 
            "autobiography for sale in\n" + 
            "the book store i just flip\n" + 
            "to the about the author\n" + 
            "section. I'm like,\n" + 
            "[Done, next!]",

            //                          /
            "When they were naming the\n" + 
            "animals, somebody got lazy:\n" + 
            "anteater?\n" + 
            "What's it doing?\n" + 
            "It's eating ants. DONE!",

            //                          /
            "I was in my friends garage,\n" + 
            "and he had; a kite, a yo-yo,\n" + 
            "and a boomerang.\n" + 
            "I was like [Dude, you have\n" + 
            "abandonment issues]",

            //                          /
            "I'd like to play a game\n" + 
            "where you help the people\n" + 
            "who were shot in all the\n" + 
            "other games.  I'd call it\n" + 
            "[Really Busy Hospital.]",

            //                          /
            "A life vest protects you\n" +
            "from drowning,\n" +
            "\n" +
            "a bulletproof vest protects\n" + 
            "you from getting shot,\n" +
            "\n" +
            "and a sweater vest protects\n" + 
            "you from pretty girls.",

            //                          /
            "I got some new pajamas with\n" + 
            "pockets in 'em.  Which is\n" + 
            "great, because before that,\n" + 
            "I used to have to hold\n" + 
            "stuff when I slept.",

            //                          /
            "Last week I lost my temper\n" + 
            "in my karate class. Man,\n" + 
            "I'm not doing that again\n" + 
            "until I'm a black belt.\n" + 
            "Because I can tell you\n" + 
            "there's a difference\n" + 
            "between taking karate\n" + 
            "and receiving karate.",

            //                          /
            "Cottonballs are an example\n" + 
            "of something I'd want to\n" + 
            "buy,\n" + 
            "but not have as a nickname.",

            //                          /
            "How fast does a zebra have\n" + 
            "to run before it looks\n" + 
            "gray?",

            //                          /
            "I have a time machine at\n" + 
            "home.\n" +
            "\n" +
            "It only goes forward\n" + 
            "at regular speed.",

            //                          /
            "I was in a restaurant that\n" + 
            "had a sign that said\n" + 
            "[Restrooms For Customers\n" + 
            "Only.] I thought,\n" + 
            "[it must suck to work\n" + 
            "there.]",

            //                          /
            "Automatic paper towel\n" + 
            "dispensers are a solution\n" + 
            "to something that was never\n" + 
            "a problem in the first\n" + 
            "place.",

            //                          /
            "I like fruit baskets\n" + 
            "because it gives you the\n" + 
            "ability to mail someone a\n" + 
            "piece of fruit without\n" + 
            "appearing insane.",

            //                          /
            "I bought a clock, but the\n" + 
            "big hand broke off of it...\n" +
            "\n" +
            "so I just add [ish] to\n" + 
            "every number.",

            //                          /
            "I'm writing my book in\n" + 
            "fifth person, so every\n" + 
            "sentence starts out with:\n" + 
            "[I heard fron this guy\n" + 
            "who told somebody ...]",

            //                          /
            "The boomerang is\n" + 
            "Australia's chief export.\n" +
            "And then import.",

            //                          /
            "I am a mystery wrapped in\n" + 
            "an enigma wrapped in a\n" + 
            "pita.\n" + 
            "Why the pita? That counts\n" + 
            "as another mystery.",

            //                          /
            "It is illegal to yell\n" + 
            "[fire] in a crowded\n" + 
            "theater. If there is a\n" + 
            "fire, please yell something\n" + 
            "else instead, like\n" +
            "\n" +
            "[Flames!] or\n" +
            "\n" +
            "\n" +
            "[Smoke maker!] or\n" +
            "\n" +
            "\n" +
            "[Bad hot!]",
        };
    }
}
